"""
Fileio
~~~~~~

This module contains the functions that read and write files.
"""


import re
import os
import sys

import docx
import json
import pandas

from . import helper_functions



def load_json(filepath):
    """Adds error checking around loading a json file.
    
    Args:
        filepath (str): filepath to the json file
        
    Returns:
        internal_data (dict): json read from file in a dictionary
        
    Raises:
        Exception: If file opening has a problem will raise an exception.
    """
    if os.path.exists(filepath):
        try:
            with open(filepath, "r") as f:
                internal_data = json.loads(f.read())
        except Exception as e:
            raise e

        return internal_data
    else:
        helper_functions.vprint("No such file: " + filepath)
        sys.exit()




def read_previous_publications(filepath):
    """Read in the previous publication json file.
    
    If the prev_pub option was given by the user then that filepath is used to read in the file
    and it is checked to make sure the json is a list and each value is a string. If the prev_pub
    option was not given then look for a "tracker-timestamp" directory in the current working 
    directory and if it has a publications.json file then read in that file. 
    If no previous publications are found then an empty dict is returned for prev_pubs.
    
    Args:
        filepath (str or None): path to the publications JSON to read in.
        
    Returns:
        has_previous_pubs (bool): True means that a previous publications file was found
        prev_pubs (dict): dict where keys are publication ids and values are a dict of publication attributes
    """
    
    has_previous_pubs = False
    if filepath:
        
        if filepath.lower() == "ignore":
            return False, {}
        
        prev_pubs = load_json(filepath)
        has_previous_pubs = True
                            
    else:
        dir_contents = os.listdir()
        ## find all directories matching the tracker directory structure and convert the timestamps to ints to find the largest one.
        tracker_dirs = [int(re.match(r"tracker-(\d{10})", folder).group(1)) for folder in dir_contents if re.match(r"tracker-(\d{10})", folder)]
        if len(tracker_dirs) > 0:
            tracker_dirs.sort(reverse=True)
            for latest_dir in tracker_dirs:
                prev_publication_filepath = os.path.join(os.getcwd(), "tracker-"+str(latest_dir), "publications.json")
                if os.path.exists(prev_publication_filepath):
                    prev_pubs = load_json(prev_publication_filepath)
                    has_previous_pubs = True
                    break
    
    if has_previous_pubs:                
                    
        return has_previous_pubs, prev_pubs
    
    else:
        return has_previous_pubs, {}
    
    
    
    

def save_emails_to_file(email_messages, save_dir_name):
    """Save email_messages to "emails.json" in save_dir_name in the current working directory.
    
    Args:
        email_messages (dict): keys are author names and values are the of the email
        save_dir_name (str): directory name to append to the current working directory to save the emails.json file in
    """
    
    email_save_path = os.path.join(os.getcwd(), save_dir_name, "emails.json")
    
    with open(email_save_path, 'w') as outFile:
        print(json.dumps(email_messages, indent=2, sort_keys=False), file=outFile)
        
            


def save_publications_to_file(save_dir_name, publication_dict, prev_pubs):
    """Saves the publication_dict to "publications.json" in save_dir_name in the current working directory.
    
    prev_pubs and publication_dict will be combined before saving.
    
    Args:
        save_dir_name (str): directory name to append to the current working directory to save the publications.json file in
        publication_dict (dict): dictionary with publication ids as the keys to the dict
        prev_pubs (list): List of publication ids that are publications previously found.
    """
    
    publications_save_path = os.path.join(os.getcwd(), save_dir_name, "publications.json")
    
    prev_pubs.update(publication_dict)
    with open(publications_save_path, 'w') as outFile:
        print(json.dumps(prev_pubs, indent=2, sort_keys=True), file=outFile)

        
        

def read_text_from_docx(doc_path):
    """Open docx file at doc_path and read contents into a string.
    
    Args:
        doc_path (str): path to docx file.
        
    Returns:
        (str): A string of the contents of the docx file. Each line concatenated with a newline character.
    
    Raises:
        Exception: If file opening has a problem will raise an exception.
    """
    
    ## https://stackoverflow.com/questions/25228106/how-to-extract-text-from-an-existing-docx-file-using-python-docx
    if os.path.exists(doc_path):
        try:
            document = docx.Document(doc_path)
            return u"\n".join([u"".join([r.text for r in paragraph._element.xpath(".//w:t")]) for paragraph in document.paragraphs])
        except Exception as e:
            raise e
    else:
        helper_functions.vprint("No such file: " + doc_path)
        sys.exit()



def read_text_from_txt(doc_path):
    """Open txt or csv file at doc_path and read contents into a string.
    
    Args:
        doc_path (str): path to txt or csv file.
        
    Returns:
        (str): A string of the contents of the txt or csv file. Each line concatenated with a newline character. 
    
    Raises:
        Exception: If file opening has a problem will raise an exception.
    """
    
    if os.path.exists(doc_path):
        try:
            with open(doc_path, encoding = "utf-8") as document:
                lines = document.readlines()
        except Exception as e:
            raise e
        
        return "".join(lines)
    else:
        helper_functions.vprint("No such file: " + doc_path)
        sys.exit()
    

def read_csv(doc_path):
    """Read csv into a pandas dataframe.
    
    Args:
        doc_path (str): path to the csv file to read in.
        
    Returns:
        df (DataFrame): Pandas dataframe of the csv contents.
    
    Raises:
        Exception: If file opening has a problem will raise an exception.
    """
    
    if os.path.exists(doc_path):
        try:
            df = pandas.read_csv(doc_path)
        except Exception as e:
            raise e
        
        return df
    else:
        helper_functions.vprint("No such file: " + doc_path)
        sys.exit()

         
    

def save_string_to_file(save_dir_name, file_name, text_to_save):
    """Save a string to file.
    
    Args:
        save_dir_name (str): directory in the current working directory to save the string to.
        file_name (str): string to name the file.
        text_to_save (str): the string to put in the file contents.
    """
    
    save_path = os.path.join(os.getcwd(), save_dir_name, file_name)
    
    with open(save_path, 'wb') as outFile:
        outFile.write(text_to_save.encode("utf-8"))
    
    

def save_json_to_file(save_dir_name, file_name, json_dict):
    """Saves the json_dict to file_name in save_dir_name in the current working directory.
    
    Args:
        save_dir_name (str): directory name to append to the current working directory to save the json_dict in.
        json_dict (dict or list): data to save to file.
    """
    
    save_path = os.path.join(os.getcwd(), save_dir_name, file_name)
    
    with open(save_path, 'w') as outFile:
        print(json.dumps(json_dict, indent=2, sort_keys=True), file=outFile)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sys
import re
from setuptools import setup, find_packages


if sys.argv[-1] == 'publish':
    os.system('python3 setup.py sdist')
    os.system('twine upload dist/*')
    sys.exit()


def readme():
    with open('README.rst') as readme_file:
        return readme_file.read()


def find_version():
    with open('academic_tracker/__init__.py', 'r') as fd:
        version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                            fd.read(), re.MULTILINE).group(1)
    if not version:
        raise RuntimeError('Cannot find version information')
    return version


REQUIRES = [
    "docopt >= 0.6.2",
    "pymed >= 0.8.9",
    "jsonschema >= 4.4.0",
    "habanero >= 1.0.0",
    "orcid >= 1.0.3",
    "scholarly >= 1.4.5",
    "beautifulsoup4 >= 4.9.3",
    "fuzzywuzzy >= 0.18.0",
    "python-docx >= 0.8.11",
    "pandas >= 0.24.2",
    "openpyxl >= 2.6.2",
    "requests >= 2.21.0",
]


setup(
    name='academic_tracker',
    version=find_version(),
    author='Travis Thompson',
    author_email='ptth222@gmail.com',
    description='Find publications on PubMed, Crossref, ORCID, and Google Scholar for given authors or references.',
    keywords='PubMed publications citations Crossref ORCID Google Scholar',
    license='BSD',
    url='https://github.com/MoseleyBioinformaticsLab/academic_tracker',
    packages=find_packages(exclude=['doc', 'docs', 'vignettes']),
    platforms=['any'],
    long_description=readme(),
    long_description_content_type="text/x-rst",
    install_requires=REQUIRES,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    entry_points={"console_scripts": ["academic_tracker = academic_tracker.__main__:main"]},
)


# -*- coding: utf-8 -*-
"""
Webio
~~~~~

General functions that interface with the internet.
"""

import urllib.request
import urllib.error
import email.message
import subprocess
import io
import re
import os
import shutil

import orcid
import scholarly
import habanero
import bs4
import json
import requests

from . import helper_functions


TOOL = "Academic Tracker"
DOI_URL = "https://doi.org/"

PUBLICATION_TEMPLATE = {
        "abstract": None,
        "authors": None,
        "conclusions": None,
        "copyrights": None,
        "doi": None,
        "journal": None,
        "keywords": None,
        "methods": None,
        "publication_date": {"year":None, "month":None, "day":None},
        "pubmed_id": None,
        "results": None,
        "title": None,
        "grants": None,
        "PMCID": None
   }



##TODO look into adding expanded search to orcid, would need to upgrade to 3.0.
def search_ORCID_for_ids(ORCID_key, ORCID_secret, authors_json):
    """Query ORCID with author names and get ORCID IDs.
    
    If an author already has an ORCID, or doesn't have affiliations they are skipped.
    
    Args:
        ORCID_key (str): key assigned to your registered application from ORCID.
        ORCID_secret (str): secret given to you by ORCID.
        authors_json (dict): JSON matching the Authors section of the Configuration file.
        
    Returns:
        authors_json (dict): the authors_json modified with any ORCID IDs found.
    """
    
    ## This is a workaround to get the expanded search functionality that the orcid package doesn't currently have.
    SEARCH_VERSION = "/v3.0"
    def search_replace(self, query, method, start, rows, headers, endpoint):
        url = endpoint + SEARCH_VERSION + "/expanded-search/?defType=" + method + "&q=" + query
        if start:
            url += "&start=%s" % start
        if rows:
            url += "&rows=%s" % rows

        response = requests.get(url, headers=headers, timeout=self._timeout)
        response.raise_for_status()
        
        if self.do_store_raw_response:
            self.raw_response = response
        return response.json()
    
    orcid.PublicAPI._search = search_replace
    
    api = orcid.PublicAPI(ORCID_key, ORCID_secret)
    
    search_token = api.get_search_token_from_orcid()
    
    for author, author_attributes in authors_json.items():
        
        if ("ORCID" in author_attributes and author_attributes["ORCID"]) or not "affiliations" in author_attributes:
            continue
        
        search_results = api.search(author_attributes["pubmed_name_search"], access_token=search_token)
        
        for result in search_results["expanded-result"]:
            if re.match(author_attributes["first_name"].lower() + ".*", result["given-names"].lower()) and author_attributes["last_name"].lower() == result["family-names"].lower():
                
                if any([affiliation.lower() in institution.lower() for institution in result["institution-name"] for affiliation in author_attributes["affiliations"]]):
                    authors_json[author]["ORCID"] = result["orcid-id"]
                    break


    return authors_json



def search_Google_Scholar_for_ids(authors_json):
    """Query Google Scholar with author names and get Scholar IDs.
    
    If an author already has a scholar_id, or doesn't have affiliations they are skipped.
    
    Args:
        authors_json (dict): JSON matching the Authors section of the Configuration file.
        
    Returns:
        authors_json (dict): the authors_json modified with any ORCID IDs found.
    """
    
    for author, author_attributes in authors_json.items():
        
        if ("scholar_id" in author_attributes and author_attributes["scholar_id"]) or not "affiliations" in author_attributes:
            continue
    
        search_query = scholarly.scholarly.search_author(author_attributes["pubmed_name_search"])
        
        for queried_author in search_query:
            name = queried_author["name"].split(" ")
            first_name = name[0].lower()
            last_name = name[-1].lower()
            if author_attributes["first_name"].lower() == first_name and author_attributes["last_name"].lower() == last_name:
        
                if any([affiliation.lower() in queried_author["affiliation"].lower() for affiliation in author_attributes["affiliations"]]):
                    authors_json[author]["scholar_id"] = queried_author["scholar_id"]
                    break
            
    return authors_json


            

def get_DOI_from_Crossref(title, mailto_email):
    """Search title on Crossref and try to find a DOI for it.
    
    Args:
        title (str): string of the title of the journal article to search for.
        mailto_email (str): an email address needed to search Crossref more effectively.
        
    Returns:
        doi (str): Either None or the DOI of the article title. The DOI will not be a URL.
    """
    
    doi = None
    
    cr = habanero.Crossref(ua_string = "Academic Tracker (mailto:" + mailto_email + ")")
    
    results = cr.works(query_bibliographic = title, filter = {"type":"journal-article"})
    
    for work in results["message"]["items"]:
        
        if not "title" in work or not helper_functions.is_fuzzy_match_to_list(title, work["title"]):
            continue
            
        ## Look for DOI
        if "DOI" in work:
            doi = work["DOI"].lower()
        
        ## Crossref should only have one result that matches the title, so if it got past the check at the top break.
        break
    
    return doi




def get_url_contents_as_str(url):
    """Query the url and return it's contents as a string.
    
    Args:
        url (str): the URL to query.
        
    Returns:
        (str): Either the website as a string or None if an error occurred.
    """
    
    try:
        req = urllib.request.Request(url, headers={"User-Agent": "Mozilla/5.0"})
        response = urllib.request.urlopen(req, timeout=5)
        url_str = response.read().decode("utf-8")
        response.close()
        return url_str
                
    except urllib.error.HTTPError as e:
        helper_functions.vprint(e, verbosity=1)
        helper_functions.vprint(url, verbosity=1)
        
        return None
    
    

def clean_tags_from_url(url):
    """Remove tags from webpage.
    
    Remove tags from a webpage so it looks more like what a user would see in a 
    browser.
    
    Args:
        url (str): the URL to query.
        
    Returns:
        clean_url (str): webpage contents cleaned of tags.    
    """
    
    url_str = get_url_contents_as_str(url)
    if not url_str:
        return None
    
    ## All of the python libraries return the html with newlines in seemingly 
    ## arbitrary locations, so remove them and add in some for tags that make sense.
    clean_url = url_str.replace("\n", "")
    clean_url = clean_url.replace("<br>", "\n")
    clean_url = clean_url.replace("</div>", "</div>\n")
    clean_url = clean_url.replace("</p>", "</p>\n")
    clean_url = bs4.BeautifulSoup(clean_url, "lxml").text
#    clean_url = lxml.html.fromstring(url_str).text_content()
    
    return clean_url



def send_emails(email_messages):
    """Uses sendmail to send email_messages to authors.
    
    Only works on systems with sendmail installed. 
    
    Args:
        email_messages (dict): keys are author names and values are the message
    """
    if not shutil.which("sendmail"):
        helper_functions.vprint("Warning: sendmail was not found in PATH, so no emails were sent.")
        return
    
    # build and send each message by looping over the email_messages dict
    for email_parts in email_messages["emails"]:
        msg = email.message.EmailMessage()
        msg["Subject"] = email_parts["subject"]
        msg["From"] = email_parts["from"]
        msg["To"] = email_parts["to"]
        msg["Cc"] = email_parts["cc"]
        msg.set_content(email_parts["body"])
        
        if os.path.exists(email_parts["attachment"]):
            with open(email_parts["attachment"], 'rb') as content_file:
                content = content_file.read()
                
            msg.add_attachment(content, maintype="application", subtype="vnd.ms-excel", filename=email_parts["attachment_filename"])
        else:
            msg.add_attachment(email_parts["attachment"], filename=email_parts["attachment_filename"])
        
        subprocess.run(["sendmail", "-t", "-oi"], input=msg.as_bytes())




###############
## Unused Functions
###############
        
def get_grants_from_Crossref(title, mailto_email, grants):
    """Search title on Crossref and try to find the grants associated with it.
    
    Only the grants in the grants parameter are searched for because trying to find 
    all grants associated with the article is too difficult.
    
    Args:
        title (str): string of the title of the journal article to search for.
        mailto_email (str): an email address needed to search Crossref more effectively.
        grants (list): a list of the grants to try and find for the article.
        
    Returns:
        found_grants (str): Either None or a list of grants found for the article.
    """
    
    found_grants = None
    
    cr = habanero.Crossref(ua_string = "Academic Tracker (mailto:" + mailto_email + ")")
    
    results = cr.works(query_bibliographic = title, filter = {"type":"journal-article"})
    
    for work in results["message"]["items"]:
        
        if not "title" in work or not helper_functions.is_fuzzy_match_to_list(title, work["title"]):
            continue
            
        if "funder" in work:
            ## the grant string could be in any value of the funder dict so look for it in each one.
            found_grants = {grant for funder in work["funder"] for value in funder.values() for grant in grants if grant in value}
            if found_grants:
                found_grants = list(found_grants)
            
        ## Crossref should only have one result that matches the title, so if it got past the check at the top break.
        break
    
    return found_grants



def get_redirect_url_from_doi(doi):
    """"""
    
    doi = doi.lower()
    
    if re.match(r".*http.*", doi):
        match = helper_functions.regex_match_return(r".*doi.org/(.*)", doi)
        if match:
            url = DOI_URL + "api/handles/" + match[0]
        else:
            return ""
    else:
        url = DOI_URL + "api/handles/" + doi
        
    try:
        req = urllib.request.Request(url)
        response = urllib.request.urlopen(req)
        json_response = json.loads(response.read())
        response.close()
                
    except urllib.error.HTTPError:
        helper_functions.vprint("Error trying to resolve DOI: " + doi, verbosity=1)
        return ""
        
    for value in json_response["values"]:
        if value["type"] == "URL":
            return value["data"]["value"]
        
    return ""



def scrape_url_for_DOI(url):
    """Searches url for DOI.
    
    Uses the regex "(?i).*doi:\s*([^\s]+\w).*" to look for a DOI on 
    the provided url. The DOI is visited to confirm it is a proper DOI.
    
    Args:
        url (str): url to search.
        
    Returns:
        DOI (str): string of the DOI found on the webpage. Is empty string if DOI is not found.
    """
        
    url_str = get_url_contents_as_str(url)
    if not url_str:
        return ""
            
    doi = helper_functions.regex_group_return(helper_functions.regex_search_return(r"(?i)doi:\s*(<[^>]*>)?([^\s<]+)", url_str), 1)
    
    if doi:
        
        url = get_redirect_url_from_doi(doi)
        return doi if url else ""
        
    else:
        return ""
    


def check_doi_for_grants(doi, grants):
    """Searches DOI webpage for grants.
    
    Concatenates "https://doi.org/" with the doi, visits the 
    page and looks for the given grants on that page.
    
    Args:
        doi (str): DOI for the publication.
        grants (list): list of str for each grant to look for.
        
    Returns:
        found_grants (list): list of str with each grant that was found on the page.
    """
    
    url = get_redirect_url_from_doi(doi)
    if not url:
        return set()
    
    url_str = get_url_contents_as_str(url)
    if not url_str:
        return set()
        
    return { grant for grant in grants if grant in url_str }
    



def download_pdf(pdf_url):
    """
    """
    ## test url https://realpython.com/python-tricks-sample-pdf
    try:
        req = urllib.request.Request(pdf_url, headers={"User-Agent": "Mozilla/5.0"})
        response = urllib.request.urlopen(req)
        pdf_bytes = io.BytesIO(response.read())
        response.close()
                
    except urllib.error.HTTPError as e:
        helper_functions.vprint(e, verbosity=1)
        helper_functions.vprint(pdf_url, verbosity=1)
        
        return None
            
    return pdf_bytes


    


# -*- coding: utf-8 -*-
"""
Reference Search Emails and Reports
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Functions to create emails and reports for reference_search.
"""

import re
import copy
import os

import pandas

from . import helper_functions
from . import fileio


DEFAULT_SUMMARY_TEMPLATE = "<pub_loop>Reference Line:\n\t<ref_line>\nTokenized Reference:\n\tAuthors: <tok_authors>\n\tTitle: <tok_title>\n\tPMID: <tok_PMID>\n\tDOI: <tok_DOI>\nQueried Information:\n\tDOI: <DOI>\n\tPMID: <PMID>\n\tPMCID: <PMCID>\n\tGrants: <grants>\n\n</pub_loop>"


def convert_tokenized_authors_to_str(authors):
    """Combine authors into a comma separated string.
    
    Try to do first_name last_name for each author, but if first name isn't there
    then last_name initials. ex. first_name1 last_name1, last_name2 initials2
    
    Args:
        authors (list): a list of dictionaries [{"last":last_name, "initials":initials}, {"last":last_name, "first":first_name}]
        
    Returns:
        authors_string (str): comma separated list of authors.
    """
    
    authors_string = ""
    for author in authors:
        if "first" in author:
            if author["first"]:
                authors_string += author["first"]
                authors_string += " " + author["last"] + ", " if author["last"] else ", "
            elif author["last"]:
                authors_string += author["last"] + ", "
        else:
            if author["last"]:
                authors_string += author["last"]
                authors_string += " " + author["initials"] + ", " if author["initials"] else ", "
            else:
                authors_string += author["initials"] + ", "
        
    authors_string = authors_string[:-2]
    
    if not authors_string:
        authors_string = str(None)
            
    return authors_string



simple_publication_keywords_map = {"<abstract>":"abstract",
                                    "<conclusions>":"conclusions",
                                    "<copyrights>":"copyrights",
                                    "<DOI>":"doi",
                                    "<journal>":"journal",
                                    "<keywords>":"keywords",
                                    "<methods>":"methods",
                                    "<PMID>":"pubmed_id",
                                    "<results>":"results",
                                    "<title>":"title",
                                    "<PMCID>":"PMCID",}

pub_authors_keyword_map = {"<pub_author_first>":"firstname",
                           "<pub_author_last>":"lastname",
                           "<pub_author_initials>":"initials",
                           "<pub_author_affiliations>":"affiliation"}

publication_date_keywords_map = {"<publication_year>":"year",
                                 "<publication_month>":"month",
                                 "<publication_day>":"day"}

tokenized_keywords_map = {"<tok_title>":"title", 
                          "<tok_DOI>":"DOI", 
                          "<tok_PMID>":"PMID"}
    

    
def create_report_from_template(publication_dict, is_citation_in_prev_pubs_list, tokenized_citations, template_string = DEFAULT_SUMMARY_TEMPLATE):
    """Create project report based on template_string.
    
    Loop over each publication in publication_dict and build a report based on 
    the tags in the template_string. Details about reports are in the documentation.
    
    Args:
        publication_dict (dict): keys and values match the publications JSON file.
        is_citation_in_prev_pubs_list (list): list of bools that indicate whether or not the citation at the same index in tokenized_citations is in the prev_pubs
        tokenized_citations (list): list of dicts. Matches the JSON schema for tokenized citations.
        template_string (str): string with tags indicated what information to put in the report.
        
    Returns:
        report (str): text of the created report.
    """
    
    matching_key_for_citation = [citation["pub_dict_key"] for citation in tokenized_citations]
    
    pub_template = helper_functions.regex_group_return(helper_functions.regex_match_return(r"(?s).*<pub_loop>(.*)</pub_loop>.*", template_string), 0)
    pub_author_template = helper_functions.regex_group_return(helper_functions.regex_match_return(r"(?s).*<pub_author_loop>(.*)</pub_author_loop>.*", template_string), 0)

    
    report_string = ""
    for pub_id, pub_values in publication_dict.items():
        pub_template_copy = pub_template
        
        pub_authors = ""
        for pub_author in pub_values["authors"]:
            pub_author_template_copy = pub_author_template
            
            for keyword, pub_author_key in pub_authors_keyword_map.items():
                pub_author_template_copy = pub_author_template_copy.replace(keyword, str(pub_author[pub_author_key]))
                
            pub_authors += pub_author_template_copy
            
        pub_template_copy = re.sub(r"(?s)<pub_author_loop>.*</pub_author_loop>", pub_authors, pub_template_copy)
        
        for keyword, pub_key in simple_publication_keywords_map.items():
            pub_template_copy = pub_template_copy.replace(keyword, str(pub_values[pub_key]))
            
        ## build first and last author
        first_author = str(pub_values["authors"][0]["lastname"]) + ", " + str(pub_values["authors"][0]["firstname"])
        pub_template_copy = pub_template_copy.replace("<first_author>", first_author)
        
        last_author = str(pub_values["authors"][-1]["lastname"]) + ", " + str(pub_values["authors"][-1]["firstname"])
        pub_template_copy = pub_template_copy.replace("<last_author>", last_author)
        
        authors = ", ".join([str(author["firstname"]) + " " + str(author["lastname"]) for author in pub_values["authors"]])
        pub_template_copy = pub_template_copy.replace("<authors>", authors)
        
        grants = ", ".join(pub_values["grants"]) if pub_values["grants"] else "None"
        pub_template_copy = pub_template_copy.replace("<grants>", grants)
        
        for keyword, date_key in publication_date_keywords_map.items():
            pub_template_copy = pub_template_copy.replace(keyword, str(pub_values["publication_date"][date_key]))
        
        tok_index = matching_key_for_citation.index(pub_id)
        for keyword, tok_key in tokenized_keywords_map.items():
            replacement = str(tokenized_citations[tok_index][tok_key])
            if not replacement:
                replacement = "None"
            pub_template_copy = pub_template_copy.replace(keyword, replacement)
            
        tok_authors = convert_tokenized_authors_to_str(tokenized_citations[tok_index]["authors"])
        pub_template_copy = pub_template_copy.replace("<tok_authors>", tok_authors)
        
        if tokenized_citations[tok_index]["reference_line"]:
            pretty_print = tokenized_citations[tok_index]["reference_line"].split("\n")
            pretty_print = " ".join([line.strip() for line in pretty_print])
            pub_template_copy = pub_template_copy.replace("<ref_line>", pretty_print)
        else:
            pub_template_copy = pub_template_copy.replace("<ref_line>", "N/A")
        
        if is_citation_in_prev_pubs_list:
            pub_template_copy = pub_template_copy.replace("<is_in_comparison_file>", str(is_citation_in_prev_pubs_list[tok_index]))
        else:
            pub_template_copy = pub_template_copy.replace("<is_in_comparison_file>", "N/A")
        
        report_string += pub_template_copy
        
    report = re.sub(r"(?s)<pub_loop>.*</pub_loop>", report_string, template_string)

    return report



def create_tabular_report(publication_dict, config_dict, is_citation_in_prev_pubs_list, tokenized_citations, save_dir_name):
    """Create a pandas DataFrame and save it as Excel or CSV.
    
    Args:
        publication_dict (dict): keys and values match the publications JSON file.
        config_dict (dict): keys and values match the project tracking configuration JSON file.
        is_citation_in_prev_pubs_list (list): list of bools that indicate whether or not the citation at the same index in tokenized_citations is in the prev_pubs
        tokenized_citations (list): list of dicts. Matches the JSON schema for tokenized citations.
        save_dir_name (str): directory to save the report in.
        
    Returns:
        report (str): Either the text of the report if csv or a relative filepath to where the Excel file is saved.
        filename (str): Filename of the report. Made have had an .xlsx added to the end.
    """
    
    rows = []
    row_template = copy.deepcopy(config_dict["summary_report"]["columns"])
    
    matching_key_for_citation = [citation["pub_dict_key"] for citation in tokenized_citations]
    
    separator = config_dict["summary_report"]["separator"] if "separator" in config_dict["summary_report"] else ","
    
    sort = config_dict["summary_report"]["sort"] if "sort" in config_dict["summary_report"] else []
    
    if "column_order" in config_dict["summary_report"]:
        column_order = config_dict["summary_report"]["column_order"]
    else:
        column_order = list(row_template.keys())
    
    file_format = config_dict["summary_report"]["file_format"] if "file_format" in config_dict["summary_report"] else "csv"
        
    if "filename" in config_dict["summary_report"]:
        filename = config_dict["summary_report"]["filename"]
    else:
        filename = "summary_report.csv" if file_format == "csv" else "summary_report.xlsx"
    
    row_string = "".join(row_template.values()) 
    if any([pub_author_keyword in row_string for pub_author_keyword in pub_authors_keyword_map.keys()]):
        has_pub_author_keywords = True
    else:
        has_pub_author_keywords = False
    
    
    for pub, pub_values in publication_dict.items():
        tok_index = matching_key_for_citation.index(pub)
        is_citation_in_prev_pubs = is_citation_in_prev_pubs_list[tok_index] if is_citation_in_prev_pubs_list else None
        
        if has_pub_author_keywords:
            for pub_author in publication_dict[pub]["authors"]:
                
                rows.append(replace_keywords(row_template, publication_dict, pub, tokenized_citations[tok_index], is_citation_in_prev_pubs, pub_author))
                
        else:
            rows.append(replace_keywords(row_template, publication_dict, pub, tokenized_citations[tok_index], is_citation_in_prev_pubs))
            
    
    report = ""            
    if rows:
        df = pandas.DataFrame(rows)
        if sort:
            df = df.sort_values(by=sort)
        df = df.drop_duplicates()
        df = df[column_order]
        
        if file_format == "csv":
            report = df.to_csv(index=False, sep=separator, line_terminator="\n")
            fileio.save_string_to_file(save_dir_name, filename, report)
        else:
            ## If the file extension isn't .xlsx then there will be an error, so force it.
            extension = os.path.splitext(filename)[1][1:].lower()
            if not extension == "xlsx":
                filename += ".xlsx"
            
            report = os.path.join(save_dir_name, filename)
            df.to_excel(report, index=False)
            
    return report, filename




def replace_keywords(template, publication_dict, pub, tokenized_citation, is_citation_in_prev_pubs, pub_author={}):
    """Replace keywords in the values of the template dictionary.
    
    Args:
        template (dict): keys are column names and values are what the elements of the column should be.
        publication_dict (dict): keys and values match the publications JSON file.
        pub (str): the key to the pub in publication_dict.
        tokenized_citation (dict): The tokenized citation from the reference for the publication.
        is_citation_in_prev_pubs (bool or None): Whether this publication is in the previous publications or not. If None then it isn't applicable.
        pub_author (dict): The author in pub.
        
    Returns:
        template_copy (dict): template with the keywords replaced in its values.
    """
    
    template_copy = copy.deepcopy(template)
    
    for key in template_copy:
                
        for keyword, pub_key in simple_publication_keywords_map.items():
            template_copy[key] = template_copy[key].replace(keyword, str(publication_dict[pub][pub_key]))
            
        ## build first and last author
        first_author = str(publication_dict[pub]["authors"][0]["lastname"]) + ", " + str(publication_dict[pub]["authors"][0]["firstname"])
        template_copy[key] = template_copy[key].replace("<first_author>", first_author)
        
        last_author = str(publication_dict[pub]["authors"][-1]["lastname"]) + ", " + str(publication_dict[pub]["authors"][-1]["firstname"])
        template_copy[key] = template_copy[key].replace("<last_author>", last_author)
        
        authors = ", ".join([str(author["firstname"]) + " " + str(author["lastname"]) for author in publication_dict[pub]["authors"]])
        template_copy[key] = template_copy[key].replace("<authors>", authors)
        
        grants = ", ".join(publication_dict[pub]["grants"]) if publication_dict[pub]["grants"] else "None Found"
        template_copy[key] = template_copy[key].replace("<grants>", grants)
        
        for keyword, date_key in publication_date_keywords_map.items():
            template_copy[key] = template_copy[key].replace(keyword, str(publication_dict[pub]["publication_date"][date_key]))
        
        ## Pub authors keywords
        if pub_author:
            for keyword, pub_author_key in pub_authors_keyword_map.items():
                template_copy[key] = template_copy[key].replace(keyword, str(pub_author[pub_author_key]))
                
        ## tokenized keywords
        for keyword, tok_key in tokenized_keywords_map.items():
            replacement = str(tokenized_citation[tok_key])
            if not replacement:
                replacement = "None"
            template_copy[key] = template_copy[key].replace(keyword, replacement)
            
        tok_authors = convert_tokenized_authors_to_str(tokenized_citation["authors"])
        template_copy[key] = template_copy[key].replace("<tok_authors>", tok_authors)
        
        if tokenized_citation["reference_line"]:
            pretty_print = tokenized_citation["reference_line"].split("\n")
            pretty_print = " ".join([line.strip() for line in pretty_print])
            template_copy[key] = template_copy[key].replace("<ref_line>", pretty_print)
        else:
            template_copy[key] = template_copy[key].replace("<ref_line>", "N/A")
        
        if type(is_citation_in_prev_pubs) == bool:
            template_copy[key] = template_copy[key].replace("<is_in_comparison_file>", str(is_citation_in_prev_pubs))
        else:
            template_copy[key] = template_copy[key].replace("<is_in_comparison_file>", "N/A")
            
    return template_copy




def create_tokenization_report(tokenized_citations):
    """Create a report that details all the information about how a reference was tokenized.
    
    Intended as a troubleshooting report.
    
    Args:
        tokenized_citations (list): list of dicts. Matches the JSON schema for tokenized citations.
        
    Returns:
        report_string (str): report text built from tokenized_citations.
    """
    
    report_string = ""
    for count, citation in enumerate(tokenized_citations):
        if tokenized_citations[count]["reference_line"]:
            pretty_print = tokenized_citations[count]["reference_line"].split("\n")
            pretty_print = " ".join([line.strip() for line in pretty_print])
            report_string += "Reference Line: \n\t" + pretty_print + "\n"
        else:
            report_string += "Reference Line: \n\tN/A\n"
        
        report_string += "Tokenized Reference: \n\tAuthors: " + convert_tokenized_authors_to_str(citation["authors"])
        report_string += "\n\tTitle: " + citation["title"] if citation["title"] else "\n\tTitle: None"
        report_string += "\n\tPMID: " + str(citation["PMID"]) if citation["PMID"] else "\n\tPMID: None"
        report_string += "\n\tDOI: " + citation["DOI"] if citation["DOI"] else "\n\tDOI: None"
        report_string += "\n\n"
        
    return report_string



###############
## Unused
###############
    

def create_reference_search_diagnostic(publication_dict, is_citation_in_prev_pubs_list, tokenized_citations):
    """"""
    
    report_string = ""
    for count, citation in enumerate(tokenized_citations):
        if tokenized_citations[count]["reference_line"]:
            pretty_print = tokenized_citations[count]["reference_line"].split("\n")
            pretty_print = " ".join([line.strip() for line in pretty_print])
            report_string += "Reference Line: " + pretty_print + "\n"
        
        report_string += "Tokenized Reference: \n\tAuthors: " + convert_tokenized_authors_to_str(citation["authors"]) + " \n\tTitle: " + citation["title"]
        if citation["PMID"]:
            report_string += " \n\tPMID: " + str(citation["PMID"])
        if citation["DOI"]:
            report_string += " \n\tDOI: " + citation["DOI"]
        report_string += "\n"
        
        if tokenized_citations[count]["pub_dict_key"]:
            doi = publication_dict[tokenized_citations[count]["pub_dict_key"]]["doi"]
            pmid = publication_dict[tokenized_citations[count]["pub_dict_key"]]["pubmed_id"]
            pmcid = publication_dict[tokenized_citations[count]["pub_dict_key"]]["PMCID"]
            if publication_dict[tokenized_citations[count]["pub_dict_key"]]["grants"]:
                grants = ", ".join(publication_dict[tokenized_citations[count]["pub_dict_key"]]["grants"])
        
                
        if not doi:
            doi = "Not Found"
        if not pmid:
            pmid = "Not Found"
        if not pmcid:
            pmcid = "Not Found"
        if not grants:
            grants = "None Found"
        
        report_string += "Queried Information: \n\tDOI: " + doi + \
                         " \n\tPMID: " + pmid + \
                         " \n\tPMCID: " + pmcid +\
                         " \n\tGrants: " + grants
        if is_citation_in_prev_pubs_list:
            report_string += " \n\tIs In Comparison File: " + str(is_citation_in_prev_pubs_list[count])
        
        report_string += "\n\n\n"
        
    return report_string





# -*- coding: utf-8 -*-
"""
User Input Checking
~~~~~~~~~~~~~~~~~~~

Functions that check the user input for errors.
"""


import sys
import re
import copy

import jsonschema

from . import tracker_schema
from . import helper_functions





def tracker_validate(instance, schema, pattern_messages={}, cls=None, *args, **kwargs):
    """Wrapper around jsonchema.validate to give better error messages.
    
    Args:
        instance (dict): JSON as a dict to validate
        schema (dict): JSON schema as a dict to validate instance against
        pattern_messages (dict): if the instance has a ValidationError of the pattern type then look up the attribute that failed the pattern in this dict and see if there is a custom message
        
    Raises:
        jsonshcema.ValidationError: If an unexpected jsonschema error happens this is raised rather than a system exit.
    """
    

    try:
        jsonschema.validate(instance=instance, schema=schema, cls=cls, *args, **kwargs)
    except jsonschema.ValidationError as e:
        ## code to easily see the contents of the error for building a better message.
#        for key, value in e._contents().items():
#            print(key, value)
#            print()
        
        message = "ValidationError: An error was found in the " + schema["title"] + ". \n"
        custom_message = ""
        
        if e.validator == "minProperties":
            message += "The " + schema["title"] + " cannot be empty."
        elif e.validator == "required":
            required_property = re.match(r"(\'.*\')", e.message).group(1)
            if len(e.relative_path) == 0:
                message += "The required property " + required_property + " is missing."
            else:
                message += "The entry " + "[%s]" % "][".join(repr(index) for index in e.relative_path) + " is missing the required property " + required_property + "."
        elif e.validator == "dependencies":
            message += "The entry " + "[%s]" % "][".join(repr(index) for index in e.relative_path) + " is missing a dependent property.\n"
            message += e.message
        elif e.validator == "dependentRequired":
            message += "The entry " + "[%s]" % "][".join(repr(index) for index in e.relative_path) + " is missing a dependent property.\n"
            message += e.message
        elif e.validator == "minLength":
            custom_message = " cannot be an empty string."
        elif e.validator == "maxLength":
            custom_message = " is too long."
        elif e.validator == "minItems":
            custom_message = " cannot be empty."
        elif e.validator == "type":
            if type(e.validator_value) == list:
                custom_message = " is not any of the allowed types: ["
                for allowed_type in e.validator_value:
                    custom_message += "\'" + allowed_type + "\', "
                custom_message = custom_message[:-2]
                custom_message += "]."
            else:
                custom_message = " is not of type \"" + e.validator_value + "\"."
        elif e.validator == "enum":
            custom_message = " is not one of [" + "%s" % ", ".join(repr(index) for index in e.validator_value) + "]"
        elif e.validator == "format":
            custom_message = " is not a valid " + e.validator_value + "."
        elif e.validator == "pattern" and e.relative_path[-1] in pattern_messages:
            custom_message = pattern_messages[e.relative_path[-1]]
        else:
            raise e
        
        
        if custom_message:
            message = message + "The value for " + "[%s]" % "][".join(repr(index) for index in e.relative_path) + custom_message
        print(message)
        sys.exit()




def cli_inputs_check(args):
    """Run input checking on the CLI inputs.
    
    Uses jsonschema to validate the inputs.
            
    Args:
        args (dict): dict from docopt.
    """
    
#    list_args = ["--grants", "--affiliations", "--cc_email"]
#    
#    for arg in list_args:
#        if args[arg]:
#            args[arg] = args[arg].split(",")
#            
#    int_args = ["--cutoff_year"]
#    
#    for arg in int_args:
#        if args[arg]:
#            try:
#                args[arg] = int(args[arg])
#            except:
#                pass
    
    tracker_validate(instance=args, schema=tracker_schema.cli_schema, format_checker=jsonschema.FormatChecker())







def config_file_check(config_json, no_ORCID, no_GoogleScholar, no_Crossref):
    """Check that the configuration JSON file is as expected.
    
    The validational jsonschema is in the tracker_schema module. 
    
    Args:
        config_json (dict): dict with the same structure as the configuration JSON file
    """
    
    schema = copy.deepcopy(tracker_schema.config_schema)
    if no_ORCID:
        del schema["properties"]["ORCID_search"]
        schema["required"].remove("ORCID_search")
    if no_Crossref and no_GoogleScholar:
        del schema["properties"]["Crossref_search"]
        schema["required"].remove("Crossref_search")
    
    pattern_messages = {"ORCID":" is not a valid ORCID. It must match the regex \d{4}-\d{4}-\d{4}-\d{3}[0,1,2,3,4,5,6,7,8,9,X]"}
    tracker_validate(instance=config_json, schema=schema, pattern_messages=pattern_messages, format_checker=jsonschema.FormatChecker())
    
    ## Make sure sort and column_order only have values in columns for any report.
    attributes_to_check = ["sort", "column_order"]
    for attribute in attributes_to_check:
        if "summary_report" in config_json and "columns" in config_json["summary_report"] and attribute in config_json["summary_report"]:
            names_not_in_columns = [name for name in config_json["summary_report"][attribute] if not name in config_json["summary_report"]["columns"]]
            if names_not_in_columns:
                helper_functions.vprint("ValidationError: The \"" + attribute + "\" attribute for the summary_report has values that are not column names in \"columns\".")
                helper_functions.vprint("The following names in \"" + attribute + "\" could not be matched to a column in \"columns\":\n\n" + "\n".join(names_not_in_columns))
                sys.exit()   
                
            if attribute == "column_order":
                if len(config_json["summary_report"]["column_order"]) != len(config_json["summary_report"]["columns"]):
                    helper_functions.vprint("ValidationError: The \"column_order\" attribute for the summary_report does not have all of the column names in \"columns\". Every column in \"columns\" must be in \"column_order\".")
                    sys.exit() 
            
            
    report_keys = ["collaborator_report", "project_report"]
    for project, project_attributes in config_json["project_descriptions"].items():
        for report_key in report_keys:
            for attribute in attributes_to_check:
                if report_key in project_attributes and "columns" in project_attributes[report_key] and attribute in project_attributes[report_key]:
                    names_not_in_columns = [name for name in project_attributes[report_key][attribute] if not name in project_attributes[report_key]["columns"]]
                    if names_not_in_columns:
                        helper_functions.vprint("ValidationError: The \"" + attribute + "\" attribute for the " + report_key + " in project " + project + " has values that are not column names in \"columns\".")
                        helper_functions.vprint("The following names in \"" + attribute + "\" could not be matched to a column in \"columns\":\n\n" + "\n".join(names_not_in_columns))
                        sys.exit()
                        
                    if attribute == "column_order":
                        if len(project_attributes[report_key]["column_order"]) != len(project_attributes[report_key]["columns"]):
                            helper_functions.vprint("ValidationError: The \"column_order\" attribute for the " + report_key + " in project " + project + " does not have all of the column names in \"columns\". Every column in \"columns\" must be in \"column_order\".")
                            sys.exit() 
                                    
    
    for author, author_attributes in config_json["Authors"].items():
        for report_key in report_keys:
            for attribute in attributes_to_check:
                if report_key in author_attributes and "columns" in author_attributes[report_key] and attribute in author_attributes[report_key]:
                    names_not_in_columns = [name for name in author_attributes[report_key][attribute] if not name in author_attributes[report_key]["columns"]]
                    if names_not_in_columns:
                        helper_functions.vprint("ValidationError: The \"" + attribute + "\" attribute for the " + report_key + " for author " + author + " has values that are not column names in \"columns\".")
                        helper_functions.vprint("The following names in \"" + attribute + "\" could not be matched to a column in \"columns\":\n\n" + "\n".join(names_not_in_columns))
                        sys.exit()
                        
                    if attribute == "column_order":
                        if len(author_attributes[report_key]["column_order"]) != len(author_attributes[report_key]["columns"]):
                            helper_functions.vprint("ValidationError: The \"column_order\" attribute for the " + report_key + " for author " + author + " does not have all of the column names in \"columns\". Every column in \"columns\" must be in \"column_order\".")
                            sys.exit() 
                    
        
            

def ref_config_file_check(config_json, no_Crossref):
    """Check that the configuration JSON file is as expected.
    
    The validational jsonschema is in the tracker_schema module.    
    
    Args:
        config_json (dict): dict with a truncated structure of the configuration JSON file
    """
    
    schema = copy.deepcopy(tracker_schema.ref_config_schema)
    if no_Crossref:
        del schema["properties"]["Crossref_search"]
        schema["required"].remove("Crossref_search")
    
    tracker_validate(instance=config_json, schema=schema, format_checker=jsonschema.FormatChecker())

    ## Make sure sort and column_order only have values in columns for any report.
    attributes_to_check = ["sort", "column_order"]
    for attribute in attributes_to_check:
        if "summary_report" in config_json and "columns" in config_json["summary_report"] and attribute in config_json["summary_report"]:
            names_not_in_columns = [name for name in config_json["summary_report"][attribute] if not name in config_json["summary_report"]["columns"]]
            if names_not_in_columns:
                helper_functions.vprint("ValidationError: The \"" + attribute + "\" attribute for the summary_report has values that are not column names in \"columns\".")
                helper_functions.vprint("The following names in \"" + attribute + "\" could not be matched to a column in \"columns\":\n\n" + "\n".join(names_not_in_columns))
                sys.exit()    
                
            if attribute == "column_order":
                if len(config_json["summary_report"]["column_order"]) != len(config_json["summary_report"]["columns"]):
                    helper_functions.vprint("ValidationError: The \"column_order\" attribute for the summary_report does not have all of the column names in \"columns\". Every column in \"columns\" must be in \"column_order\".")
                    sys.exit() 
        



def prev_pubs_file_check(prev_pubs):
    """Run input checking on prev_pubs dict.
    
    The validational jsonschema is in the tracker_schema module. 
    
    Args:
        prev_pubs (dict): dict with the same structure as the previous publications JSON file.
    """
    
    tracker_validate(instance=prev_pubs, schema=tracker_schema.publications_schema, format_checker=jsonschema.FormatChecker())
    


def tok_reference_check(tok_ref):
    """Run input checking on tok_ref dict.
    
    The validational jsonschema is in the tracker_schema module. 
            
    Args:
        tok_ref (dict): dict with the same structure as the tokenized reference JSON file.
    """
    
    tracker_validate(instance=tok_ref, schema=tracker_schema.tok_schema, format_checker=jsonschema.FormatChecker())

